import pytest
from django.db import models
from django.db.models import ExpressionWrapper, F, Q, Value, IntegerField
from django.db.models.functions import Lower
from django.db.models.lookups import Exact, GreaterThan

from dj_hybrid.expander import expand_query, Not, And, EmptyQuery
from dj_hybrid.expression_wrapper.wrap import wrap
from dj_hybrid.tests.utils import are_equal

models.CharField.register_lookup(Lower)

pytestmark = pytest.mark.django_db(transaction=True)


class FakeModel(models.Model):  # type: ignore
    char_field = models.CharField(max_length=150, default="")
    int_field = models.IntegerField(default=0)
    float_field = models.FloatField(default=0.)
    m2o_field = models.ForeignKey('self', on_delete=models.CASCADE, null=True)


def test_lookup_comparison() -> None:
    q1 = Exact(Lower(
        ExpressionWrapper(F('char_field'), output_field=models.CharField())
    ), Value(''))
    q2 = Exact(Lower(ExpressionWrapper(F('char_field'), output_field=models.CharField())), Value(''))
    q3 = GreaterThan(
        ExpressionWrapper(F('int_field'), output_field=models.IntegerField()),
        Value(2)
    )
    q4 = Exact(Lower(
        ExpressionWrapper(F('char_field'), output_field=models.TextField())
    ), Value(''))

    assert are_equal(q1, q2)
    assert not are_equal(q1, q3)
    assert not are_equal(q1, q4)


def test_char_expression() -> None:
    expected = Exact(Lower(ExpressionWrapper(F('char_field'), output_field=models.CharField())), Value(''))
    query_1 = Q(char_field__lower='')
    query_2 = Q(char_field__lower__exact='')

    expanded_1 = expand_query(FakeModel, query_1)
    expanded_2 = expand_query(FakeModel, query_2)

    assert are_equal(expected, expanded_1)
    assert are_equal(expected, expanded_2)


def test_nested_query():
    expected = Exact(
        ExpressionWrapper(F('char_field'), output_field=models.CharField()),
        Value(''))
    query = Q(Q(char_field=''))
    expanded = expand_query(FakeModel, query)
    assert are_equal(expected, expanded)


def test_negated_query():
    expected = Not(Exact(
        ExpressionWrapper(F('char_field'), output_field=models.CharField()),
        Value('')
    ))
    query_1 = ~Q(char_field='')
    query_2 = Q(~Q(char_field=''))
    expanded_1 = expand_query(FakeModel, query_1)
    expanded_2 = expand_query(FakeModel, query_2)
    assert are_equal(expected, expanded_1)
    assert are_equal(expected, expanded_2)


def test_and_query():
    expected = And(
        Exact(ExpressionWrapper(F('char_field'), output_field=models.CharField()), Value('')),
        Exact(ExpressionWrapper(F('int_field'), output_field=models.IntegerField()), Value(1)),
    )
    queries = [
        # Q(char_field='', int_field=1),  # Order is not guaranteed!
        Q(Q(char_field=''), int_field=1),
        Q(Q(char_field=''), Q(int_field=1)),
        Q(char_field='') & Q(int_field=1),
    ]
    for query in queries:
        expanded = expand_query(FakeModel, query)
        assert are_equal(expected, expanded)


def test_empty_query():
    expected = EmptyQuery()
    query = Q()
    expanded = expand_query(FakeModel, query)
    assert are_equal(expected, expanded)
    assert wrap(expanded).as_python(None)


def test_negated_empty_query():
    expected = EmptyQuery()
    query = ~Q()
    expanded = expand_query(FakeModel, query)
    assert are_equal(expected, expanded)
    assert wrap(expanded).as_python(None)


def test_combining_empty_query():
    # when Django combines an empty with an empty, it will only output one.
    expected = EmptyQuery()
    query = Q() & Q()
    expanded = expand_query(FakeModel, query)
    assert are_equal(expected, expanded)
    assert wrap(expanded).as_python(None)


def test_combining_empty_with_query():
    expected = Exact(ExpressionWrapper(F('int_field'), output_field=IntegerField()), Value(1))
    query = Q() & Q(int_field=1)
    expanded = expand_query(FakeModel, query)
    assert are_equal(expected, expanded)
    assert wrap(expanded).as_python(dict(int_field=1))


def test_multiple_empty_children():
    expected = EmptyQuery()
    query = Q(Q(), Q())
    expanded = expand_query(FakeModel, query)
    assert are_equal(expected, expanded)
    assert wrap(expanded).as_python(None)


def test_adding_empty_as_a_child():
    expected = Exact(ExpressionWrapper(F('int_field'), output_field=IntegerField()), Value(1))
    query = Q(Q(), Q(int_field=1))
    expanded = expand_query(FakeModel, query)
    assert are_equal(expected, expanded)
    assert wrap(expanded).as_python(dict(int_field=1))


def test_empty_in_empty():
    expected = EmptyQuery()
    query = Q(Q(Q(Q())))
    expanded = expand_query(FakeModel, query)
    assert are_equal(expected, expanded)
    assert expanded.as_python(None)


def test_negated_empty_as_a_sibling():
    expected = Exact(ExpressionWrapper(F('int_field'), output_field=IntegerField()), Value(1))
    query = Q(~Q(), Q(int_field=1))
    expanded = expand_query(FakeModel, query)
    assert are_equal(expected, expanded)
    assert wrap(expanded).as_python(dict(int_field=1))



import factory
from typing import ClassVar, Union, Mapping, Any, Type

import pytest

from dj_hybrid.expression_wrapper.wrap import wrap

from django.db.models import Lookup, Q, F, Expression, Model


ExpressionType = Union[Expression, F, Q, Lookup]


_UNSET = object()


@pytest.mark.django_db(transaction=True)
class WrapperTestBase:
    factory = None  # type: ClassVar[Type[factory.Factory]]
    model_class = None  # type: ClassVar[Type[Model]]
    fixture = {}  # type: ClassVar[Mapping[str, Any]]

    python_value = _UNSET  # type: ClassVar[Any]
    expression = None  # type: ClassVar[ExpressionType]

    def python_equivalent(self, obj: Model) -> Any:
        if self.python_value is _UNSET:
            raise NotImplementedError
        return self.python_value

    def get_expression(self) -> ExpressionType:
        return self.expression

    def get_populated_model(self) -> Model:
        if not hasattr(self, '__model_instance'):
            self.__model_instance = self.factory(**self.fixture)
        return self.__model_instance

    def get_as_python(self, model_instance):
        expression = self.get_expression()
        wrapped = wrap(expression)
        return wrapped.as_python(model_instance)

    def get_from_database(self, model_instance):
        expression = self.get_expression()
        model_class = model_instance._meta.model

        # we need to run this as an annotation, so fake it a little
        annotated_instance = model_class.objects.annotate(
            _wrapped_testing=expression,
        ).get(pk=model_instance.pk)
        return annotated_instance._wrapped_testing

    def test_expression_evaluates_to_expected(self):
        model_instance = self.get_populated_model()

        expression_outcome = self.get_as_python(model_instance)
        python_outcome = self.python_equivalent(model_instance)
        assert expression_outcome == python_outcome

    def test_parity_with_database(self):
        model_instance = self.get_populated_model()
        python_outcome = self.get_as_python(model_instance)
        database_outcome = self.get_from_database(model_instance)

        assert python_outcome == database_outcome



from functools import partial
from inspect import signature
from unittest.mock import Mock

from django.db.models import F
from pytest import raises

import dj_hybrid
from dj_hybrid import decorator, hybrid_property
from dj_hybrid.decorator import HybridProperty, NamedExpression
from .utils import not_raises, are_equal


def get_some_class():
    # we wrap this so that we get a clean cache on each test
    class SomeClass:
        def __init__(self, int_field=1, char_field="hello"):
            self.int_field = int_field
            self.char_field = char_field

        @hybrid_property
        def int_field_alias(cls):
            return F('int_field')

        @dj_hybrid.property
        def char_field_alias(cls):
            return F('char_field')

        @dj_hybrid.property
        def add_2(cls):
            return cls.int_field_alias + 20
    return SomeClass


def test_interface():
    assert hybrid_property is dj_hybrid.property is HybridProperty

    # we follow an uncalled decorator pattern, return_value=None
    # essentially. @hybrid, not @hybrid()
    sig = signature(hybrid_property)
    with not_raises(TypeError):
        sig.bind(object)

    # we also support wrapping a function when defining a name
    with not_raises(TypeError):
        sig.bind(object, name="some_name")

    # but we don't support setting the name implicitly
    with raises(TypeError):
        sig.bind(object, "some_name")


def test_int_field_alias__class():
    expected = NamedExpression(F('int_field'), 'int_field_alias')
    actual = get_some_class().int_field_alias
    assert are_equal(expected, actual)


def test_int_field_alias__instance():
    expected = 1
    SomeClass = get_some_class()
    actual = SomeClass(int_field=expected).int_field_alias
    assert expected == actual


def test_caching_behaviour__class(mocker):
    mocked_named = mocker.spy(NamedExpression, '__init__')  # type: Mock
    SomeClass = get_some_class()
    first = SomeClass.int_field_alias
    second = SomeClass.int_field_alias
    assert mocked_named.call_count == 1
    assert first is second
    SomeClass.__dict__['int_field_alias'].reset_cache()
    third = SomeClass.int_field_alias
    assert mocked_named.call_count == 2
    assert third is not first
    assert are_equal(first, third)


def test_caching_behaviour__instance(mocker):
    mocked_wrap = mocker.spy(decorator, 'wrap')  # type: Mock
    SomeClass = get_some_class()
    instance = SomeClass(char_field="hello")

    descriptor = SomeClass.__dict__['char_field_alias']

    first = instance.char_field_alias
    first_cache = descriptor._cached_wrapped
    second = instance.char_field_alias
    second_cache = descriptor._cached_wrapped

    assert first is second
    assert first_cache is second_cache
    assert mocked_wrap.call_count == 1

    SomeClass.__dict__['char_field_alias'].reset_cache()
    assert not descriptor._cached_wrapped

    third = instance.char_field_alias
    third_cache = descriptor._cached_wrapped
    assert first_cache is not third_cache
    assert are_equal(first_cache, third_cache)
    assert first is third
    assert mocked_wrap.call_count == 2



from factory import RelatedFactory, SubFactory
from factory.django import DjangoModelFactory
from faker import Faker

from .models import WrapperStubModel, FTestingModel, FTestingRelatedModel

fake = Faker()


class WrapperStubFactory(DjangoModelFactory):
    class Meta:
        model = WrapperStubModel


class FTestingRelatedFactory(DjangoModelFactory):
    class Meta:
        model = FTestingRelatedModel

    int_field = fake.pyint()
    str_field = fake.pystr(max_chars=FTestingRelatedModel._meta.get_field('str_field').max_length)


class FTestingFactory(DjangoModelFactory):
    class Meta:
        model = FTestingModel

    int_field = fake.pyint()
    str_field = fake.pystr(max_chars=FTestingModel._meta.get_field('str_field').max_length)
    related = SubFactory(FTestingRelatedFactory)



import operator
from typing import Any, Callable, Tuple, Type, Union

from django.db.models import ExpressionWrapper, F, Field, FieldDoesNotExist, Model, Q, Value
from django.db.models.constants import LOOKUP_SEP
from django.db.models.lookups import Exact, Lookup
from django.db.models.options import Options
from dj_hybrid.expression_wrapper.types import Wrapable

from .expression_wrapper.wrap import wrap

Connector_T = Union['Or', 'And']
LookupOrConnector_T = Union[Lookup, Connector_T]
Expanded_T = Union[LookupOrConnector_T, 'Not']


def expand_query(model: Type[Model], query: Q) -> Wrapable:
    """
    Expand query expressions into a more consumable form

    The basic structure of a query is essentially:
      <field>[_<field>]+[_<transform>]+_<lookup>.
    note: even when absent, the lookup is always there. it defaults to `_exact`.

    We want to convert it into a form that will allow offloading logic into our
    expression wrappers.

    We do not expect the output to be consumable within Django. This is only for
    our python bindings.

    example:
    >>> Q(date_field__year__gte=2017)
    becomes:
    >>> YearGte(YearExtract(F('date_field')), 2017)

    Some quirks to be aware of:
    When combining, we return some special classes; Not and And.

    When negating, a special Or class is returned.
    These will not function correctly as Django expressions.

    When an empty query is encountered, we return a special EmptyQuery object.
    This will always return True.
    It is also
    """

    try:
        first_child = query.children[0]
    except IndexError:
        return EmptyQuery()
    if isinstance(first_child, Q):
        expanded = expand_query(model, first_child)
    else:
        expanded = expand_child(model, first_child)

    connector = get_connector(query.connector)

    for child in query.children[1:]:
        if isinstance(child, Q):
            expanded = connector(expanded, expand_query(model, child))
        else:
            expanded = connector(expanded, expand_child(model, child))

    if query.negated:
        expanded = Not(expanded)

    return expanded


def expand_child(model: Type[Model], child: Tuple[str, Any]) -> Lookup:
    arg, value = child

    if not isinstance(value, Value):
        value = Value(value)

    parts = arg.split(LOOKUP_SEP)
    opts = model._meta  # type: Options
    inner_opts = opts
    field = None  # type: Field
    pos = 0

    # we need to work out the full field path, which we will put in an F()
    for pos, part in enumerate(parts):
        if part == 'pk':
            part = inner_opts.pk.name
        try:
            field = inner_opts.get_field(part)
        except FieldDoesNotExist:
            break
        else:
            if field.is_relation:
                inner_opts = field.model._meta

    else:
        # we never broke out, which means everything resolved correctly.
        # bump pos by one so we get no remainder
        pos += 1

    if field is None:
        raise Exception("Field not found: {}".format(parts))

    field_path = LOOKUP_SEP.join(parts[:pos])
    expression = F(field_path)

    # we set lookup_expression to field as that's what we're gathering from.
    # It will be updated in parallel with `expression` later on
    lookup_expression = field
    # we need to wrap the F() so we can specify the output field.
    # It's kind of bastardised..
    expression = ExpressionWrapper(expression, output_field=field)

    remainder = parts[pos:]
    if not remainder:
        remainder = ['exact']

    # we don't try to access the last entry, as that is probably a lookup
    for part in remainder[:-1]:
        transformer = lookup_expression.get_transform(part)
        if not transformer:
            raise Exception("Invalid transform: {}".format(part))
        lookup_expression = expression = transformer(expression)

    lookup_name = remainder[-1]
    lookup_class = lookup_expression.get_lookup(lookup_name)
    if not lookup_class:
        transformer = lookup_expression.get_transform(lookup_name)
        if not transformer:
            raise Exception("invalid transform or field access: {}".format(lookup_name))
        lookup_expression = expression = transformer(expression)
        lookup_name = 'exact'
        lookup_class = lookup_expression.get_lookup(lookup_name)

    # we'd rather use isnull instead of Eq(None)
    if value.value is None and lookup_name in ('exact', 'iexact'):
        return lookup_expression.get_lookup('isnull')(expression, True)
    return lookup_class(expression, value)


def get_connector(connector_name: Union[Q.AND, Q.OR]) -> Callable[[Wrapable, Wrapable], Wrapable]:
    connector = {
        Q.AND: And,
        Q.OR: Or,
    }[connector_name]

    def guard_empty(lhs: Wrapable, rhs: Wrapable) -> Wrapable:
        nonlocal connector
        if isinstance(lhs, EmptyQuery):
            return rhs
        if isinstance(rhs, EmptyQuery):
            return lhs
        # to reduce as much exposure to EmptyQuery as possible, remove it where we know it's
        # safe to remove. Django treats it as invisible, so should we.

        return connector(lhs, rhs)
    return guard_empty


class Combineable:
    combiner = None  # type: Callable[[bool, bool], bool]

    def __init__(self, lhs: Wrapable, rhs: Wrapable) -> None:
        self.lhs, self.rhs = lhs, rhs

    def as_python(self, obj: Any) -> bool:
        wrapped_lhs = wrap(self.lhs)
        wrapped_rhs = wrap(self.rhs)
        return type(self).combiner(wrapped_lhs.as_python(obj), wrapped_rhs.as_python(obj))


class And(Combineable):
    combiner = operator.and_


class Or(Combineable):
    combiner = operator.or_


class Not:
    def __init__(self, expression: Wrapable) -> None:
        self.expression = expression

    def as_python(self, obj: Any) -> bool:
        return not wrap(self.expression).as_python(obj)


class EmptyQuery:

    @staticmethod
    def as_python(obj: Any) -> bool:
        return True

    def __eq__(self, other: Any) -> bool:
        return type(self) is type(other)



from .utils import not_raises


class CustomException(Exception):
    pass


def test_not_utils():
    exception = CustomException

    def force_raise(): raise exception
    def force_no_raise(): pass

    # no failed assertions will happen if nothing raises
    with not_raises(exception):
        force_no_raise()

    # we need to catch the assertion error. If we get one, great.
    # if we don't, we need to throw our own, as it didn't catch the correct exception
    try:
        with not_raises(exception):
            force_raise()
    except AssertionError as e:
        pass  # we're all good here!
    except exception as e:
        assert not e





import pytest
from django.db import models
from django.db.models import Case, Q, Value, When

from dj_hybrid.expression_wrapper.wrap import wrap


class ControlFlowModel(models.Model):
    char_field = models.CharField(max_length=150, default="")
    int_field = models.IntegerField(default=0)
    float_field = models.FloatField(default=0.)
    m2o_field = models.ForeignKey('self', on_delete=models.CASCADE, null=True)


@pytest.mark.django_db(transaction=True)
def test_query():
    expected = True
    query = Q(int_field__gt=20)
    wrapped = wrap(query)
    instance = ControlFlowModel(int_field=50)

    assert expected == wrapped.as_python(instance)


@pytest.mark.django_db(transaction=True)
def test_case():
    expected = "Woot!"
    expression = Case(
        When(int_field__lt=20, then=Value("got 20!")),
        When(int_field=23, then=Value("Woot!")),
    )
    wrapped = wrap(expression)
    instance = ControlFlowModel(int_field=23)

    assert expected == wrapped.as_python(instance)



import operator
import re
import statistics
from datetime import date, datetime
from typing import (
    Any,
    AnyStr,
    Callable,
    Container,
    Dict,
    Generator,
    Generic,
    Iterable,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import django
from django.core.exceptions import FieldError
from django.db.models import (
    Aggregate,
    Avg,
    Count,
    F,
    Field,
    Func,
    Lookup,
    Max,
    Min,
    Model,
    Q,
    StdDev,
    Sum,
    Value,
    Variance,
    When,
)
from django.db.models.expressions import (
    Case,
    Col,
    Combinable,
    CombinedExpression,
    DurationValue,
    ExpressionWrapper as DjangoExpressionWrapper,
    Random,
)
from django.db.models.functions import Cast, Coalesce, Concat, ConcatPair, Greatest, Least, Length, Lower, Now, Upper
from django.db.models.lookups import (
    Contains,
    EndsWith,
    Exact,
    GreaterThan,
    GreaterThanOrEqual,
    IContains,
    IEndsWith,
    IExact,
    In,
    IntegerGreaterThanOrEqual,
    IntegerLessThan,
    IRegex,
    IsNull,
    IStartsWith,
    LessThan,
    LessThanOrEqual,
    Range,
    Regex,
    StartsWith,
    Transform,
)
from django.utils import timezone
from django.utils.crypto import random

from dj_hybrid.expander import expand_query
from dj_hybrid.resolve import get_resolver
from dj_hybrid.types import SupportsPython, SupportsPythonComparison
from dj_hybrid.utils import cached_property

from .base import ExpressionWrapper, FakeQuery
from .registry import register
from .wrap import wrap


"""
There are some main types of expressions in Django:
  - Expressions
    This is the base. It acts as a template for Values and Refs
  - Func
    This can be any function call. It defines how many arguments there are, and
    what function is to be called.
    - Aggregate
      This is a subclass of Func. It takes an arbitrary amount of arguments, however
      there is nothing really special about it.
    - Transform
      This is a subclass of Func. They take only one argument, and are commonly used
      in queries.
  - Lookup
    These are not technically expressions.
    They take 2 expressions, and should return True or False.
    The first expression should point at a field, the latter can be anything.
  - References
    These point at fields. They're not technically Expressions, but they behave
    the same.
  - Queries
    A query (or a `Q` object) is actually just syntactical sugar. It wraps a field
    reference, transforms, and a lookup.
    It is used in filters, as well as control flows.
  - Control Flows
    Django uses `Case` and `When` to supply control flows. this is essentially chained
    If statements.
"""


class OutputFieldMixin:
    def to_value(self, value: Any) -> Any:
        resolved_expression = self.resolved_expression  # type: ignore
        try:
            field = resolved_expression.field  # type: Field
        except FieldError:
            pass  # no output_field defined
        else:
            if field:
                # TODO: decide if it's better to use get_db_converters here
                value = field.to_python(value)
        return value


@register(Value)
@register(DurationValue)
class ValueWrapper(ExpressionWrapper[Union[Value, DurationValue]], OutputFieldMixin):

    def as_python(self, obj: Any) -> Any:
        value = self.resolved_expression.value
        return self.to_value(value)


@register(CombinedExpression)
class CombinedExpressionWrapper(ExpressionWrapper[CombinedExpression], OutputFieldMixin):

    _connectors = {
        Combinable.ADD: operator.add,
        Combinable.SUB: operator.sub,
        Combinable.MUL: operator.mul,
        Combinable.DIV: operator.truediv,
        Combinable.MOD: operator.mod,
        Combinable.BITAND: operator.and_,
        Combinable.BITOR: operator.or_,
        Combinable.BITLEFTSHIFT: operator.lshift,
        Combinable.BITRIGHTSHIFT: operator.rshift,
    }  # type: Dict[str, Callable[[Any, Any], Any]]

    def as_python(self, obj: Any) -> Any:

        lhs_wrapped = wrap(self.resolved_expression.lhs)
        rhs_wrapped = wrap(self.resolved_expression.rhs)
        lhs = lhs_wrapped.as_python(obj)
        rhs = rhs_wrapped.as_python(obj)
        op = self._get_operator()
        value = op(lhs, rhs)
        return self.to_value(value)

    def _get_operator(self) -> Callable[[Any, Any], Any]:
        connector = self.resolved_expression.connector  # type: str
        op = self._connectors[connector]
        return op


@register(Col)
class ColWrapper(ExpressionWrapper[Col]):

    def as_python(self, obj: Any) -> Any:
        resolver = get_resolver(obj)
        resolved = resolver.resolve(self.resolved_expression.alias)

        # This behaviour might not be right, but everything I've seen suggests it..
        # We need to turn a model instance into its PK value.
        if isinstance(resolved, Model):
            return resolved.pk
        return resolved


@register(F)
def f_resolver(expression: F) -> ColWrapper:
    # F doesn't behave the same as a normal expression.
    # It essentially acts an alias for either Col or Ref.
    # This will probably cause some broken parts down the road,
    # however let's actually get this running first!

    expression = expression.resolve_expression(
        FakeQuery(),
    )
    return ColWrapper(expression)


@register(Random)
class RandomWrapper(ExpressionWrapper[Random], OutputFieldMixin):

    def as_python(self, obj: Any) -> float:
        return cast(float, self.to_value(
            self.consistent_random,
        ))

    @cached_property
    def consistent_random(self) -> float:
        return cast(float, random.random())


@register(DjangoExpressionWrapper)
class ExpressionWrapperWrapper(ExpressionWrapper[DjangoExpressionWrapper], OutputFieldMixin):

    def as_python(self, obj: Any) -> Any:
        wrapped = wrap(self.expression.expression)
        value = wrapped.as_python(obj)
        return self.to_value(value)


T_Func = TypeVar('T_Func', bound=Func)


class FuncWrapper(ExpressionWrapper[Func], OutputFieldMixin, Generic[T_Func]):
    op = None  # type: Callable

    def as_python(self, obj: Any) -> Any:
        input_values = self.get_source_values(obj)
        output_value = type(self).op(*input_values)
        return self.to_value(output_value)

    def get_source_values(self, obj: Any) -> Generator[Any, None, None]:
        for expression in self.resolved_expression.source_expressions:
            wrapped = wrap(expression)
            yield wrapped.as_python(obj)


T_Aggregate = TypeVar('T_Aggregate', bound=Aggregate)


class AggregateWrapper(FuncWrapper[Aggregate], Generic[T_Aggregate]):
    op = None  # type: Callable[[Iterable[Any]], Any]


@register(Avg)
class AvgWrapper(AggregateWrapper[Avg]):
    op = statistics.mean


@register(Count)
class CountWrapper(AggregateWrapper[Count]):
    op = len


@register(Max)
class MaxWrapper(AggregateWrapper[Count]):
    op = max


@register(Min)
class MinWrapper(AggregateWrapper[Min]):
    op = min


@register(StdDev)
class StdDevWrapper(AggregateWrapper[StdDev]):
    op = statistics.stdev


@register(Sum)
class SumWrapper(AggregateWrapper[Sum]):
    op = sum


@register(Variance)
class VarianceWrapper(AggregateWrapper[Variance]):
    op = statistics.variance


Cast_T = TypeVar('Cast_T')


@register(Cast)
class CastWrapper(FuncWrapper[Cast]):
    @staticmethod
    def op(value: Cast_T) -> Cast_T:
        return value


Coalesce_T = TypeVar('Coalesce_T')


@register(Coalesce)
class CoalesceWrapper(FuncWrapper[Coalesce]):
    @staticmethod
    def op(*values: Coalesce_T) -> Optional[Coalesce_T]:
        return next(
            (value for value in values if value is not None),
            None
        )


@register(ConcatPair)
@register(Concat)
class ConcatPairWrapper(FuncWrapper[Union[Concat, ConcatPair]]):
    @staticmethod
    def op(*values: str) -> str:
        return ''.join(str(v) for v in values)


@register(Greatest)
class GreatestWrapper(FuncWrapper[Greatest]):
    op = max


@register(Least)
class LeastWrapper(FuncWrapper[Least]):
    op = min


@register(Length)
class LengthWrapper(FuncWrapper[Length]):
    op = len


T_Transform = TypeVar('T_Transform', bound=Transform)


class TransformWrapper(FuncWrapper[Transform], Generic[T_Transform]):
    op = None  # type: Callable[[Any], Any]


@register(Now)
class NowWrapper(ExpressionWrapper[Now], OutputFieldMixin):
    def as_python(self, obj: Any) -> datetime:
        return cast(datetime, self.to_value(self.consistent_now))

    @cached_property
    def consistent_now(self) -> datetime:
        return cast(datetime, timezone.now())


@register(Lower)
class LowerWrapper(TransformWrapper[Lower]):
    @staticmethod
    def op(value: AnyStr) -> AnyStr:
        if value:
            return value.lower()
        return value


@register(Upper)
class UpperWrapper(TransformWrapper[Upper]):
    @staticmethod
    def op(value: AnyStr) -> AnyStr:
        if value:
            return value.upper()
        return value


try:
    from django.db.models.functions import StrIndex
except ImportError:
    pass
else:
    @register(StrIndex)
    class StrIndexWrapper(FuncWrapper[StrIndex]):
        @staticmethod
        def op(string: AnyStr, lookup: AnyStr) -> int:
            try:
                value = string.index(lookup)
            except ValueError:
                value = 0
            return value


T_Lookup = TypeVar('T_Lookup', bound=Lookup)


class LookupWrapper(ExpressionWrapper[Lookup], Generic[T_Lookup]):
    op = None  # type: Callable[[Any, Any], bool]

    def as_python(self, obj: Any) -> bool:
        lhs_wrapped = wrap(self.resolved_expression.lhs)
        rhs_wrapped = self.get_wrapped_rhs()
        lhs_value = lhs_wrapped.as_python(obj)
        rhs_value = rhs_wrapped.as_python(obj)
        return type(self).op(lhs_value, rhs_value)

    def get_wrapped_rhs(self) -> SupportsPython:
        rhs = self.resolved_expression.rhs
        for transform in self.resolved_expression.bilateral_transforms:
            rhs = transform(rhs)
        return wrap(rhs)


@register(Exact)
class ExactWrapper(LookupWrapper[Exact]):
    op = operator.eq


@register(IExact)
class IExactWrapper(LookupWrapper[IExact]):
    @staticmethod
    def op(lhs: AnyStr, rhs: AnyStr) -> bool:
        if lhs and rhs:
            return lhs.lower() == rhs.lower()
        return lhs == rhs


# TODO: python doesn't like comparing different types. investigate.

@register(GreaterThan)
class GreaterThanWrapper(LookupWrapper[GreaterThan], Generic[T_Lookup]):
    op = operator.gt


@register(GreaterThanOrEqual)
@register(IntegerGreaterThanOrEqual)
class GreaterThanOrEqualWrapper(LookupWrapper[Union[GreaterThanOrEqual, IntegerGreaterThanOrEqual]], Generic[T_Lookup]):
    op = operator.ge


@register(LessThan)
@register(IntegerLessThan)
class LessThanWrapper(LookupWrapper[Union[LessThan, IntegerLessThan]], Generic[T_Lookup]):
    op = operator.lt


@register(LessThanOrEqual)
class LessThanOrEqualWrapper(LookupWrapper[LessThanOrEqual], Generic[T_Lookup]):
    op = operator.le


@register(In)
class InWrapper(LookupWrapper[In]):
    @staticmethod
    def op(lhs: Any, rhs: Container) -> bool:
        # TODO: support querysets?
        return lhs in rhs


@register(Contains)
class ContainsWrapper(LookupWrapper[Contains]):
    op = operator.contains


@register(IContains)
class IContainsWrapper(LookupWrapper[IContains]):
    @staticmethod
    def op(lhs: AnyStr, rhs: AnyStr) -> bool:
        if lhs and rhs:
            return rhs.lower() in lhs.lower()
        return rhs in lhs


@register(StartsWith)
class StartsWithWrapper(LookupWrapper[StartsWith]):
    op = str.startswith


@register(IStartsWith)
class IStartsWithWrapper(LookupWrapper[IStartsWith]):
    @staticmethod
    def op(lhs: AnyStr, rhs: AnyStr) -> bool:
        if lhs and rhs:
            return lhs.lower().startswith(rhs.lower())
        # unsure on this..
        return lhs.startswith(rhs)


@register(EndsWith)
class EndsWithWrapper(LookupWrapper[EndsWith]):
    op = str.endswith


@register(IEndsWith)
class IEndsWithWrapper(LookupWrapper[IEndsWith]):
    @staticmethod
    def op(lhs: AnyStr, rhs: AnyStr) -> bool:
        return lhs.lower().endswith(rhs.lower())


Rangeable_T = TypeVar('Rangeable_T', int, date)


@register(Range)
class RangeWrapper(LookupWrapper[Range]):

    @staticmethod
    def op(lhs: Rangeable_T, rhs: Tuple[Rangeable_T, Rangeable_T]) -> bool:
        return rhs[0] >= lhs <= rhs[1]


@register(IsNull)
class IsNullWrapper(LookupWrapper[IsNull]):
    @staticmethod
    def op(lhs: Any, wants_null: bool) -> bool:
        if wants_null:
            return lhs is None
        return lhs is not None


@register(Regex)
class RegexWrapper(LookupWrapper[Regex], Generic[T_Lookup]):
    re_flags = 0

    @classmethod
    def op(cls, lhs: str, rhs: str) -> bool:
        re_rhs = re.compile(rhs, flags=cls.re_flags)
        return bool(re_rhs.search(lhs))


@register(IRegex)
class IRegexWrapper(RegexWrapper[IRegex]):
    re_flags = re.IGNORECASE


@register(Q)
class QWrapper(ExpressionWrapper[Q]):

    def as_python(self, obj: Model) -> bool:
        expanded_query = expand_query(obj._meta.model, self.expression)
        wrapped = wrap(expanded_query)
        return cast(SupportsPythonComparison, wrapped).as_python(obj)


class ConditionNotMet(Exception):
    pass


@register(Case)
class CaseWrapper(ExpressionWrapper[Case], OutputFieldMixin):

    def as_python(self, obj: Any) -> Any:
        statement = self.resolved_expression
        for case in statement.cases:  # type: When
            wrapped_case = wrap(case)
            try:
                value = wrapped_case.as_python(obj)
                break
            except ConditionNotMet:
                pass
        else:
            value = wrap(statement.default).as_python(obj)
        return self.to_value(value)


@register(When)
class WhenWrapper(ExpressionWrapper[When]):

    def as_python(self, obj: Any) -> Any:
        statement = self.resolved_expression
        wrapped_condition = wrap(statement.condition)
        if wrapped_condition.as_python(obj):
            return wrap(statement.result).as_python(obj)
        raise ConditionNotMet


if django.VERSION < (2,):
    from django.db.models.lookups import DecimalGreaterThan, DecimalGreaterThanOrEqual, DecimalLessThan, DecimalLessThanOrEqual

    register(DecimalGreaterThan, GreaterThanWrapper[DecimalGreaterThan])
    register(DecimalGreaterThanOrEqual, GreaterThanOrEqualWrapper[DecimalGreaterThanOrEqual])
    register(DecimalLessThan, LessThanWrapper[DecimalLessThan])
    register(DecimalLessThanOrEqual, LessThanOrEqualWrapper[DecimalGreaterThanOrEqual])



from typing import Any

from typing_extensions import Protocol, runtime


@runtime
class SupportsPython(Protocol):
    def as_python(self, obj: Any) -> Any:
        ...


class SupportsPythonComparison(SupportsPython, Protocol):
    def as_python(self, obj: Any) -> bool:
        ...



from typing import Any

import pytest
from django.db.models import Avg, CharField, ExpressionWrapper, F, Value

from dj_hybrid.expression_wrapper.wrap import wrap


def obj(**kwargs: Any) -> object:
    return type('fakeObject', (object,), kwargs)


@pytest.mark.parametrize('expected,value', [
    ('test', Value('test')),
    (1.0, Value(1.0)),
    (str(1), Value(1, output_field=CharField())),
    (1 + 1, Value(1) + Value(1)),
    (6 % 5, Value(6) % Value(5))
])
def test_value(expected: Any, value: Any) -> None:
    wrapped_value = wrap(value)
    python_value = wrapped_value.as_python(None)
    assert python_value == expected


@pytest.mark.parametrize('field_name,value', [
    ('test1', 1.0),
    ('test2', 'value'),
])
def test_field(field_name: str, value: Any) -> None:
    fake_obj = obj(**{field_name: value})
    field = F(field_name)
    wrapped_field = wrap(field)

    output_value = wrapped_field.as_python(fake_obj)
    assert output_value == value


@pytest.mark.parametrize('expected,obj,expression', [
    (5, obj(f1=3, f2=2), F('f1')+F('f2')),
    (5, obj(f1=2), F('f1') + 3),
    (8, obj(f1=5), 3 + F('f1')),
    (str(8), obj(f1=8), ExpressionWrapper(F('f1'), CharField()))
])
def test_field_combining(expected: Any, obj: Any, expression: Any) -> None:
    wrapped_expression = wrap(expression)
    output_value = wrapped_expression.as_python(obj)

    assert output_value == expected


@pytest.mark.parametrize('expected,inner_obj,expression', [
    (2, [1, 2, 3], Avg('nested')),
    (10.5, [0, 1, 3, 6], Avg('nested') + 8),
])
def test_aggregates(expected: Any, inner_obj: Any, expression: Any) -> None:
    nested_obj = obj(nested=inner_obj)
    wrapped_expression = wrap(expression)
    output_value = wrapped_expression.as_python(nested_obj)

    assert output_value == expected


@pytest.mark.parametrize('expected,obj,expression', [
    (5, dict(f1=3, f2=2), F('f1')+F('f2')),
])
def test_dictionary_resolving(expected: Any, obj: Any, expression: Any) -> None:
    wrapped_expression = wrap(expression)
    output_value = wrapped_expression.as_python(obj)

    assert output_value == expected



from inspect import Signature

from pytest import raises

from dj_hybrid.expression_wrapper import registry
from dj_hybrid.tests.utils import not_raises


def test_interface_exposed():
    assert hasattr(registry, 'register')
    assert hasattr(registry, 'registry')
    assert hasattr(registry.registry, 'register')
    assert hasattr(registry.registry, 'unregister')
    assert hasattr(registry.registry, 'get')

    assert registry.registry.register == registry.register


def test_register_signature():
    signature = Signature.from_callable(registry.register)

    with not_raises(TypeError):
        signature.bind(object(), object())
        signature.bind(object(), wrapper=object())
        signature.bind(expression=object(), wrapper=object())

    # register can behave as a decorator too
    returned_decorator = registry.register(object)
    signature = Signature.from_callable(returned_decorator)

    with not_raises(TypeError):
        signature.bind(object())


def test_unregister_signature():
    signature = Signature.from_callable(registry.registry.unregister)

    with not_raises(TypeError):
        signature.bind(object())


def test_get_signature():
    signature = Signature.from_callable(registry.registry.get)

    with not_raises(TypeError):
        signature.bind(object())


def test_lifecycle():
    key = object()
    val = object()

    registry.registry.register(key, val)
    assert registry.registry.get(key) is val

    assert registry.registry.unregister(key) is None
    with raises(KeyError):
        registry.registry.get(key)

    # and the same with the decorated approach
    registry.registry.register(key)(val)
    assert registry.registry.get(key) is val

    assert registry.registry.unregister(key) is None
    with raises(KeyError):
        registry.registry.get(key)


def test_double_register():
    key = object()
    val = object()

    registry.registry.register(key, val)

    with raises(ValueError, match="Already in registry"):
        registry.registry.register(key, val)

    registry.registry.unregister(key)



from typing import TYPE_CHECKING, Any, Generic, Tuple, TypeVar, cast

from dj_hybrid.utils import cached_property

from .types import Wrapable, Wrapper

if TYPE_CHECKING:
    from django.db.models import Q
    from django.db.models.expressions import Col


T_Q = TypeVar('T_Q', bound='Q')
T_Wrapable = TypeVar('T_Wrapable', bound=Wrapable)


class FakeQuery:
    @classmethod
    def resolve_ref(cls, name: str, *_: Any, **__: Any) -> 'Col':
        from django.db.models.expressions import Col
        # We need to do some faking of the ref resolution.
        # This essentially enables us to have a bit more complete
        # workings of F().
        return Col(name, None)

    @classmethod
    def _add_q(cls, node: T_Q, *_: Any, **__: Any) -> Tuple[T_Q, None]:
        return node, None

    @classmethod
    def promote_joins(cls, *_: Any, **__: Any) -> None:
        pass


class ExpressionWrapper(Wrapper, Generic[T_Wrapable]):
    def __init__(self, expression: T_Wrapable) -> None:
        super().__init__(expression)
        self.expression = expression

    @cached_property
    def resolved_expression(self) -> T_Wrapable:
        # ok, this is a bit of a lie. This can infact return any Wrapable, however
        # for most cases, it will be of the same type as to what we're dealing with.
        if hasattr(self.expression, 'resolve_expression'):
            return cast(T_Wrapable, self.expression.resolve_expression(FakeQuery))
        return self.expression



SECRET_KEY = 'fake-key'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
    }
}

INSTALLED_APPS = [
    'dj_hybrid',
    'dj_hybrid.tests.expression_wrapper.wrapper',
]



from typing import Type, Union

from dj_hybrid.types import SupportsPython

from .registry import registry
from .types import TypeWrapperOrProxy, Wrapable


def wrap(expression: Wrapable) -> SupportsPython:
    """Wrap an expression so we can control how each one works within python

    Django essentially builds a chain of expressions, which is then used to build
    a chain of SQL expressions.
    We're going to use this underlying structure, and wrap the expressions so we can
    define how an expression behaves in the python domain.

    :param expression:
    :return:
    """
    if isinstance(expression, SupportsPython):
        return expression

    wrapper = get_wrapper(expression)
    return wrapper(expression)


def get_wrapper(expression: Union[Wrapable, Type[Wrapable]]) -> TypeWrapperOrProxy:
    if not isinstance(expression, type):
        expression = type(expression)
    return registry.get(expression)



from django.db import models


class WrapperStubModel(models.Model):
    pass


class FTestingModel(models.Model):
    int_field = models.IntegerField()
    str_field = models.CharField(max_length=150)
    related = models.ForeignKey('FTestingRelatedModel', on_delete=models.CASCADE, null=True)


class FTestingRelatedModel(models.Model):
    int_field = models.IntegerField()
    str_field = models.CharField(max_length=150)



from typing import Callable, Dict, Optional, Type, TypeVar, Union, overload

from .types import TypeWrapperOrProxy, Wrapable

Registrable = Type[Wrapable]
T_TypeWrapperOrProxy = TypeVar('T_TypeWrapperOrProxy', bound=TypeWrapperOrProxy)


class Registry:
    def __init__(self) -> None:
        self.registry = {}  # type: Dict[Type[Wrapable], TypeWrapperOrProxy]

    @overload
    def register(self, expression: Registrable, wrapper: T_TypeWrapperOrProxy) -> T_TypeWrapperOrProxy:
        ...

    @overload
    def register(
        self,
        expression: Registrable,
        wrapper: None = None
    ) -> Callable[[T_TypeWrapperOrProxy], T_TypeWrapperOrProxy]:
        ...

    def register(  # type: ignore  # these are defined in overload, but it's causing an undue conflict
        self,
        expression: Registrable,
        wrapper: Optional[T_TypeWrapperOrProxy] = None
    ) -> Union[T_TypeWrapperOrProxy, Callable[[T_TypeWrapperOrProxy], T_TypeWrapperOrProxy]]:

        if wrapper is None:
            def outer(actual_wrapper: T_TypeWrapperOrProxy) -> T_TypeWrapperOrProxy:
                nonlocal expression, self
                return self._register(expression, actual_wrapper)
            return outer

        return self._register(expression, wrapper)

    def _register(self, expression: Registrable, wrapper: T_TypeWrapperOrProxy) -> T_TypeWrapperOrProxy:
        if expression in self.registry:
            raise ValueError("Already in registry")
        self.registry[expression] = wrapper
        return wrapper

    def unregister(self, expression: Registrable) -> None:
        self.registry.pop(expression, None)

    def get(self, expression: Registrable) -> TypeWrapperOrProxy:
        return self.registry[expression]


registry = Registry()
register = registry.register



from typing import Any, Callable, Optional, Type, TypeVar, Union, cast, overload

from django.db.models import ExpressionWrapper, Expression, F
from .expression_wrapper.types import Wrapable
from .expression_wrapper.wrap import wrap
from .types import SupportsPython

T = TypeVar('T')
V_Class = TypeVar('V_Class', bound=Wrapable)
HybridMethodType = Callable[[Type[T]], V_Class]


class HybridProperty(classmethod):
    __slots__ = (
        'func',
        'name',
        '_cached_expression',
        '_cached_wrapped',
    )

    def __init__(self,
                 func: HybridMethodType,
                 *,
                 name: Optional[str] = None) -> None:
        super().__init__(func)
        self.func = func
        self.name = name or func.__name__
        self._cached_expression = None  # type: Optional[Wrapable]
        self._cached_wrapped = None  # type: Optional[SupportsPython]

    @overload
    def __get__(self, instance: T, owner: Optional[Type[T]] = None) -> Any:
        ...

    @overload
    def __get__(self, instance: None, owner: Type[T]) -> V_Class:
        ...

    def __get__(self, instance: Optional[T], owner: Optional[Type[T]] = None) -> Union[Any, V_Class]:
        if instance is None:
            return self.class_method_behaviour(cast(Type[T], owner))
        return self.instance_method_behaviour(instance)

    def reset_cache(self) -> None:
        self._cached_expression = None
        self._cached_wrapped = None

    def __del__(self) -> None:
        self.reset_cache()

    def __set_name__(self, owner: Type[T], name: str) -> None:
        self.name = name

    def class_method_behaviour(self, owner: Type[T]) -> V_Class:
        if self._cached_expression is None:
            self._cached_expression = NamedExpression(self.func(owner), self.name)
        return cast(V_Class, self._cached_expression)

    def instance_method_behaviour(self, instance: T) -> Any:
        if self._cached_wrapped is None:
            self._cached_wrapped = wrap(
                self.func(type(instance))
            )
        return self._cached_wrapped.as_python(instance)


class NamedExpression(ExpressionWrapper):  # type: ignore
    def __init__(self, expression: Union[Expression, F], default_alias: str) -> None:
        super().__init__(expression, None)
        self.default_alias = default_alias



from django.db.models import F

from .base import WrapperTestBase
from .factory import FTestingFactory
from .models import FTestingModel


class Base(WrapperTestBase):
    model_class = FTestingModel
    factory = FTestingFactory


class TestAccessInt(Base):
    expression = F('int_field')
    python_value = 24
    fixture = dict(
        int_field=python_value,
    )


class TestAccessStr(Base):
    expression = F('str_field')
    python_value = 'hello'
    fixture = dict(
        str_field=python_value,
    )


class TestAccessRelation(Base):
    expression = F('related')
    python_value = 1
    fixture = dict(
        related__pk=python_value,
    )


class TestAccessRelatedInt(Base):
    expression = F('related__int_field')
    python_value = 26
    fixture = dict(
        related__int_field=python_value,
    )


class TestCombining(Base):
    expression = F('int_field') + 24
    python_value = 50
    fixture = dict(
        int_field=26
    )



import datetime

import pytest

from django.db.models import Value, DateField, IntegerField, CharField, ExpressionWrapper, NullBooleanField

from .base import WrapperTestBase
from .factory import WrapperStubFactory
from .models import WrapperStubModel


class ValueTestBase(WrapperTestBase):
    model_class = WrapperStubModel
    factory = WrapperStubFactory


class TestInt(ValueTestBase):
    expression = Value(28, output_field=IntegerField())
    python_value = 28


class TestStr(ValueTestBase):
    expression = Value("some string", output_field=CharField())
    python_value = "some string"


class TestCombine(ValueTestBase):
    expression = Value(28, output_field=IntegerField()) + 2
    python_value = 30


class TestDurationExplicit(ValueTestBase):
    expression = ExpressionWrapper(
        Value(datetime.date(2018, 1, 1), output_field=DateField()) + datetime.timedelta(days=1),
        output_field=DateField(),
    )
    python_value = datetime.date(2018, 1, 2)


class TestNull(ValueTestBase):
    expression = Value(None, output_field=NullBooleanField())
    python_value = None


class TestMultipleCombines(ValueTestBase):
    expression = (Value(20, output_field=IntegerField()) + 10) + 20
    python_value = 50


class TestCasting(ValueTestBase):
    expression = Value("24", output_field=IntegerField())
    python_value = 24



from collections import Iterable
from contextlib import contextmanager
from functools import singledispatch
from itertools import combinations

from typing import Union, Type, Any, Sequence, Callable, TypeVar

import django
from django.db.models import Lookup, Expression, Field
from django.db.models.expressions import F
from django.utils.functional import cached_property

from dj_hybrid.expander import Not, Combineable
from dj_hybrid.expression_wrapper.base import ExpressionWrapper

ExceptionType = Union[BaseException, Type[BaseException]]


T = TypeVar('T')


_UNSET = object()


@contextmanager
def not_raises(exception: ExceptionType) -> None:
    try:
        yield
    except exception as e:
        raise AssertionError("Did raise: {}".format(e))


@singledispatch
def are_equal(*vals: T) -> bool:
    return all(
        v1 == v2
        for v1, v2 in combinations(vals, 2)
    )


@are_equal.register(list)
def iterable_equal(*vals: T) -> bool:
    return all(
        type(v1) is type(v2)
        and len(v1) == len(v2)
        and all(
            are_equal(i1, i2)
            for i1, i2 in zip(v1, v2)
        )
        for v1, v2 in combinations(vals, 2)
    )


def compare_factory(*parameters: Sequence[str]) -> Callable[..., bool]:
    def comparer(*vals):
        # type: (*T) -> bool
        nonlocal parameters
        return all(
            bool(
                type(v1) is type(v2)
                and all(
                    are_equal(
                        getattr(v1, p, _UNSET),
                        getattr(v2, p, _UNSET),
                    )
                    for p in parameters
                )
            )
            for v1, v2 in combinations(vals, 2)
        )
    return comparer


are_equal.register(Combineable, compare_factory('lhs', 'rhs'))
are_equal.register(Lookup, compare_factory('lhs', 'rhs', 'bilateral_transforms'))
are_equal.register(Not, compare_factory('expression'))


def compare_dicts(*vals: T):
    # This isn't strictly safe, however for our usecase, it's fine.
    # We need to not compare anything that's cached, as the other may also be cached
    cached_properties = set(k for k, v in type(vals[0]).__dict__.items() if isinstance(v, cached_property))
    comparer = compare_factory(*(set(vals[0].__dict__) - cached_properties))
    return all(
        type(v1) is type(v2)
        and len(set(v1.__dict__) - cached_properties) == len(set(v2.__dict__) - cached_properties)
        and comparer(v1, v2)
        for v1, v2 in combinations(vals, 2)
    )


are_equal.register(ExpressionWrapper, compare_dicts)


if django.VERSION < (2,):
    are_equal.register(F, compare_factory('name'))

    @are_equal.register(Expression)
    def compare_expression(*vals: T) -> bool:
        compare_output_field = compare_factory('_output_field')
        return all(
            type(v1) is type(v2)
            and compare_output_field(v1, v2)
            and compare_dicts(v1, v2)
            for v1, v2 in combinations(vals, 2)
        )

    @are_equal.register(Field)
    def compare_field(*vals: T) -> bool:
        return all(
            type(v1) is type(v2)
            and v1.deconstruct()[1] == v2.deconstruct()[1]
            for v1, v2 in combinations(vals, 2)
        )



from setuptools import setup

tests_require = [
    "pytest",
    "pytest-cov",
    "pytest-django",
    "pytest-factoryboy",
    "pytest-mock",
]


setup(
    tests_require=tests_require,
    extras_require={
        'test': tests_require,
    }
)



from abc import abstractmethod
from operator import attrgetter
from typing import Any, Mapping

from django.db.models import Model
from django.db.models.constants import LOOKUP_SEP

from .utils import nested_itemgetter

_notset = object()


class IResolver:
    __slots__ = (
    )

    @abstractmethod
    def resolve(self, path: str, default: Any = _notset) -> Any:
        ...


class AttributeResolver(IResolver):
    __slots__ = (
        'doc',
    )

    def __init__(self, doc: Any) -> None:
        self.doc = doc

    def resolve(self, path: str, default: Any = _notset) -> Any:
        parts = path.split(LOOKUP_SEP)
        attribute_accessor = attrgetter('.'.join(parts))
        return attribute_accessor(self.doc)


class DictResolver(IResolver):
    __slots__ = (
        'doc',
    )

    def __init__(self, doc: Mapping[str, Any]) -> None:
        self.doc = doc

    def resolve(self, path: str, default: Any = _notset) -> Any:
        parts = path.split(LOOKUP_SEP)
        item_accessor = nested_itemgetter('.'.join(parts))
        return item_accessor(self.doc)


class DjangoResolver(AttributeResolver):
    pass


def get_resolver(doc: Any) -> IResolver:
    if isinstance(doc, Model):
        return DjangoResolver(doc)
    elif isinstance(doc, Mapping):
        return DictResolver(doc)
    else:
        return AttributeResolver(doc)



from typing import Any, Callable, Generic, Mapping, Optional, Type, TypeVar, Union, overload

from django.utils.functional import cached_property as django_cached_property

ObjectStructure = Mapping[str, Any]
NestedItemCallable = Callable[[ObjectStructure], Any]


def nested_itemgetter(item: str) -> NestedItemCallable:
    def inner(obj: ObjectStructure) -> Any:
        return resolve_item(obj, item)
    return inner


def resolve_item(obj: ObjectStructure, item: str) -> Any:
    for key in item.split('.'):
        obj = obj[key]
    return obj


T = TypeVar('T')
R = TypeVar('R')


class cached_property(django_cached_property, Generic[T, R]):  # type: ignore
    def __init__(self, func: Callable[[T], R], name: Optional[str] = None) -> None:
        super().__init__(func, name=name)

    @overload
    def __get__(self, instance: None, cls: Optional[Type[T]] = None) -> 'cached_property[T, R]':
        ...

    @overload
    def __get__(self, instance: T, cls: Optional[Type[T]] = None) -> R:
        ...

    def __get__(self, instance: Optional[T], cls: Optional[Type[T]] = None) -> Union['cached_property[T, R]', R]:
        return super().__get__(instance, cls)  # type: ignore



from inspect import Signature

from dj_hybrid.expression_wrapper import wrap
from dj_hybrid.expression_wrapper.registry import registry
from dj_hybrid.tests.utils import not_raises


class FakeSupportsPython:
    def as_python(self, obj):
        ...


class FakeWrapper:
    def __init__(self, expression):
        self.expression = expression


def test_wrap_signature():
    signature = Signature.from_callable(wrap.wrap)
    with not_raises(TypeError):
        signature.bind(object())
        signature.bind(expression=object())


def test_wrap_wraps(mocker):
    mocker.patch('dj_hybrid.expression_wrapper.wrap.registry.get', return_value=FakeWrapper)

    obj_to_wrap = object()
    wrapped = wrap.wrap(obj_to_wrap)
    assert isinstance(wrapped, FakeWrapper)
    assert wrapped.expression is obj_to_wrap


def test_wrap_ignores_python_supported():
    obj_to_wrap = FakeSupportsPython()
    wrapped = wrap.wrap(obj_to_wrap)
    assert wrapped is obj_to_wrap


def test_get_wrapper():
    ObjToRegister = type('ObjToRegister', (object,), {})
    registry.register(ObjToRegister, FakeWrapper)

    assert wrap.get_wrapper(ObjToRegister) is FakeWrapper
    assert wrap.get_wrapper(ObjToRegister()) is FakeWrapper

    registry.unregister(ObjToRegister)



from typing import TYPE_CHECKING, Callable, Type, Union

from typing_extensions import Protocol

from dj_hybrid.types import SupportsPython

if TYPE_CHECKING:
    from django.db.models import Expression, F, Q, Lookup


Wrapable = Union['Expression', 'F', 'Q', 'Lookup', SupportsPython]


class Wrapper(SupportsPython, Protocol):
    def __init__(self, expression: Wrapable) -> None:
        ...


TypeWrapperOrProxy = Union[Type[Wrapper], Callable[[Wrapable], Wrapper]]
